from ophyd import (
    ADComponent as ADCpt,
    Device,
    DeviceStatus,
)

from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO

from ophyd_devices.interfaces.base_classes.psi_detector_base import PSIDetectorBase, CustomDetectorMixin

from bec_lib import bec_logger, messages
from bec_lib.endpoints import MessageEndpoints

logger = bec_logger.logger

DETECTOR_TIMEOUT = 5

class PhoenixTriggerError(Exception):
    """Base class for exceptions in this module."""


class PhoenixTriggerTimeoutError(XMAPError):
    """Raised when the PhoenixTrigger does not respond in time."""


class PhoenixTriggerDetectorState(enum.IntEnum):
    """Detector states for XMAP detector"""

    DONE = 0
    ACQUIRING = 1




class PhoenixTriggerSetup(CustomDetectorMixin):
    """
    This defines the trigger setup.


    """

    def __init__(self, *args, parent:Device = None, **kwargs):
        super().__init__(*args, parent=parent, **kwargs)
        self._counter = 0

    def on_stage(self):
        exposure_time  = self.parent.scaninfo.exp_time
        num_points = self.parent.scaninfo.num_points

        # camera acquisition parameters
        self.parent.cam.array_counter.put(0)
        if self.parent.scaninfo.scan_type == 'step':
            self.parent.cam.acquire_time.put(exposure_time)
            self.parent.cam.num_images.put(1)
            self.parent.cam.image_mode.put(0)  # Single
            self.parent.cam.trigger_mode.put(0) # auto
        else:
            # In flyscan, the exp_time is the time between two triggers,
            # which minus 15% is used as the acquisition time.
            self.parent.cam.acquire_time.put(exposure_time * 0.85)
            self.parent.cam.num_images.put(num_points)
            self.parent.cam.image_mode.put(1)  # Multiple
            self.parent.cam.trigger_mode.put(1) # trigger
            self.parent.cam.acquire.put(1, wait=False) # arm

        # file writer
        self.parent.hdf.lazy_open.put(1)
        self.parent.hdf.num_capture.put(num_points)
        self.parent.hdf.file_write_mode.put(2) # Stream
        self.parent.hdf.capture.put(1, wait=False)
        self.parent.hdf.enable.put(1) # enable plugin

        # roi statistics to collect signal and background in a timeseries
        self.parent.roistat.enable.put(1)
        self.parent.roistat.ts_num_points.put(num_points)
        self.parent.roistat.ts_control.put(0, wait=False) # Erase/Start

        logger.success('XXXX stage XXXX')

    def on_trigger(self):
        self.parent.cam.acquire.put(1, wait=False)
        logger.success('XXXX trigger XXXX')

        return self.wait_with_status(
            [(self.parent.cam.acquire.get, 0)],
            self.parent.scaninfo.exp_time + DETECTOR_TIMEOUT,
            all_signals=True
        )

    def on_complete(self):
        status = DeviceStatus(self.parent)

        if self.parent.scaninfo.scan_type == 'step':
            timeout = DETECTOR_TIMEOUT
        else:
            timeout = self.parent.scaninfo.exp_time * self.parent.scaninfo.num_points + DETECTOR_TIMEOUT
        logger.success('XXXX %s XXXX' % self.parent.roistat.ts_acquiring.get())
        success = self.wait_for_signals(
            [
                (self.parent.cam.acquire.get, 0),
                (self.parent.hdf.capture.get, 0),
                (self.parent.roistat.ts_acquiring.get, 'Done')
            ],
            timeout,
            check_stopped=True,
            all_signals=True
        )

        # publish file location
        self.parent.filepath.put(self.parent.hdf.full_file_name.get())
        self.publish_file_location(done=True, successful=success)

        # publish timeseries data
        metadata = self.parent.scaninfo.scan_msg.metadata
        metadata.update({"async_update": "append", "num_lines": self.parent.roistat.ts_current_point.get()})
        msg = messages.DeviceMessage(
            signals={
                self.parent.roistat.roi1.name_.get(): {
                    'value': self.parent.roistat.roi1.ts_total.get(),
                },
                self.parent.roistat.roi2.name_.get(): {
                    'value': self.parent.roistat.roi2.ts_total.get(),
                },
            },
            metadata=self.parent.scaninfo.scan_msg.metadata
        )
        self.parent.connector.xadd(
            topic=MessageEndpoints.device_async_readback(
                scan_id=self.parent.scaninfo.scan_id, device=self.parent.name
            ),
            msg_dict={"data": msg},
            expire=1800,
        )

        logger.success('XXXX complete %d XXXX' % success)
        if success:
            status.set_finished()
        else:
            status.set_exception(TimeoutError())
        return status

    def on_stop(self):
        logger.success('XXXX stop XXXX')
        self.parent.cam.acquire.put(0)
        self.parent.hdf.capture.put(0)
        self.parent.roistat.ts_control.put(2)

    def on_unstage(self):
        self.parent.cam.acquire.put(0)
        self.parent.hdf.capture.put(0)
        self.parent.roistat.ts_control.put(2)
        logger.success('XXXX unstage XXXX')


class EigerROIStatPlugin(ROIStatPlugin):
    roi1 = ADCpt(ROIStatNPlugin, '1:')
    roi2 = ADCpt(ROIStatNPlugin, '2:')

class PhoenixTrigger(PSIDetectorBase):

    """
    Parent class: PSIDetectorBase

    class attributes:
        custom_prepare_cls (XMAPSetup)        : Custom detector setup class for cSAXS,
                                                  inherits from CustomDetectorMixin
                                                  in __init__ of PSIDetecor bases
                                                  class is initialized
                                                  self.custom_prepare = self.custom_prepare_cls(parent=self, **kwargs)
        PSIDetectorBase.set_min_readout (float) : Minimum readout time for the detector
        dxp (EpicsDXPXMAP)                      : DXP parameters for XMAP detector
        mca (EpicsMCARecord)                    : MCA parameters for XMAP detector
        hdf5 (XMAPHDF5Plugins)                  : HDF5 parameters for XMAP detector
        MIN_READOUT (float)                     : Minimum readout time for the detector


    The class PhoenixTrigger is the class to be called via yaml configuration file
    the input arguments are defined by PSIDetectorBase,
    and need to be given in the yaml configuration file.
    To adress chanels such as  'X07MB-OP2:SMPL-DONE':

    use prefix   'X07MB-OP2:' in the device definition in the yaml configuration file.

    PSIDetectorBase(
        prefix='',
        *,Q
        name,
        kind=None,
        parent=None,
        device_manager=None,
        **kwargs,
    )
    Docstring:
    Abstract base class for SLS detectors

    Class attributes:
        custom_prepare_cls (object): class for custom prepare logic (BL specific)

    Args:
        prefix (str): EPICS PV prefix for component (optional)
        name (str): name of the device, as will be reported via read()
        kind (str): member of class 'ophydobj.Kind', defaults to Kind.normal
                    omitted -> readout ignored for read 'ophydobj.read()'
                    normal -> readout for read
                    config -> config parameter for 'ophydobj.read_configuration()'
                    hinted -> which attribute is readout for read
        parent (object): instance of the parent device
        device_manager (object): bec device manager
        **kwargs: keyword arguments
    File:           /data/test/x07mb-test-bec/bec_deployment/ophyd_devices/ophyd_devices/interfaces/base_classes/psi_detector_base.py
    Type:           type
    Subclasses:     SimCamera, SimMonitorAsync





    """
    #custom_prepare_cls = PhoenixTriggerSetup

    #cam = ADCpt(SLSDetectorCam, 'cam1:')


    #X07MB-OP2:START-CSMPL.. cont on / off

    # X07MB-OP2:SMPL.. take single sample
    #X07MB-OP2:INTR-COUNT.. counter run up
    #X07MB-OP2:TOTAL-CYCLES .. cycles set
    #X07MB-OP2:SMPL-DONE

    QUESTION HOW does ADCpt kno the EPICS prefix??????

    #image = ADCpt(ImagePlugin, 'image1:')
    #roi1 = ADCpt(ROIPlugin, 'ROI1:')
    #roi2 = ADCpt(ROIPlugin, 'ROI2:')
    #stats1 = ADCpt(StatsPlugin, 'Stats1:')
    #stats2 = ADCpt(StatsPlugin, 'Stats2:')
    roistat = ADCpt(EigerROIStatPlugin, 'ROIStat1:')
    #roistat1 = ADCpt(ROIStatNPlugin, 'ROIStat1:1:')
    #roistat2 = ADCpt(ROIStatNPlugin, 'ROIStat1:2:')
    hdf = ADCpt(HDF5Plugin, 'HDF1:')
    )
